@@ -2,7 +2,7 @@
  *   files.c  --  This file is part of GNU nano.                          *
  *                                                                        *
  *   Copyright (C) 1999-2011, 2013-2018 Free Software Foundation, Inc.    *
- *   Copyright (C) 2015-2017 Benno Schulenberg                            *
+ *   Copyright (C) 2015-2018 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
  *   it under the terms of the GNU General Public License as published    *
@@ -66,16 +66,19 @@ void make_new_buffer(void)
 
 	if (openfile == NULL) {
 		/* Make the first open file the only element in the list. */
+#ifdef ENABLE_MULTIBUFFER
 		newnode->prev = newnode;
 		newnode->next = newnode;
+#endif
 		firstfile = newnode;
 	} else {
 		/* Add the new open file after the current one in the list. */
+#ifdef ENABLE_MULTIBUFFER
 		newnode->prev = openfile;
 		newnode->next = openfile->next;
 		openfile->next->prev = newnode;
 		openfile->next = newnode;
-
+#endif
 		/* There is more than one file open: show "Close" in help lines. */
 		exitfunc->desc = close_tag;
 		more_than_one = !inhelp || more_than_one;
@@ -406,13 +409,11 @@ void stat_with_alloc(const char *filename, struct stat **pstat)
 }
 #endif /* !NANO_TINY */
 
-/* This does one of three things.  If the filename is "", just create a new
- * empty buffer.  Otherwise, read the given file into the existing buffer,
- * or into a new buffer when MULTIBUFFER is set or there is no buffer yet. */
-bool open_buffer(const char *filename, bool undoable)
+/* This does one of three things.  If the filename is "", it just creates
+ * a new empty buffer.  When the filename is not empty, it reads that file
+ * into a new buffer when requested, otherwise into the existing buffer. */
+bool open_buffer(const char *filename, bool new_buffer)
 {
-	bool new_buffer = (openfile == NULL || ISSET(MULTIBUFFER));
-		/* Whether we load into the current buffer or a new one. */
 	char *realname;
 		/* The filename after tilde expansion. */
 	FILE *f;
@@ -475,24 +476,20 @@ bool open_buffer(const char *filename, bool undoable)
 	rc = (filename[0] != '\0' && !ISSET(NOREAD_MODE)) ?
 				open_file(realname, new_buffer, inhelp, &f) : -2;
 
-	/* If we have a file, and we're loading into a new buffer, update
-	 * the filename. */
-	if (rc != -1 && new_buffer)
-		openfile->filename = mallocstrcpy(openfile->filename, realname);
-
 	/* If we have a non-new file, read it in.  Then, if the buffer has
 	 * no stat, update the stat, if applicable. */
 	if (rc > 0) {
-		read_file(f, rc, realname, undoable, new_buffer);
+		read_file(f, rc, realname, !new_buffer);
 #ifndef NANO_TINY
 		if (openfile->current_stat == NULL)
 			stat_with_alloc(realname, &openfile->current_stat);
 #endif
 	}
 
-	/* If we have a file, and we're loading into a new buffer, move back
-	 * to the beginning of the first line of the buffer. */
+	/* If we have a file, and we've loaded it into a new buffer, set
+	 * the filename and put the cursor at the start of the buffer. */
 	if (rc != -1 && new_buffer) {
+		openfile->filename = mallocstrcpy(openfile->filename, realname);
 		openfile->current = openfile->fileage;
 		openfile->current_x = 0;
 		openfile->placewewant = 0;
@@ -515,6 +512,7 @@ void replace_buffer(const char *filename)
 {
 	FILE *f;
 	int descriptor;
+	filestruct *was_cutbuffer = cutbuffer;
 
 	/* Open the file quietly. */
 	descriptor = open_file(filename, FALSE, TRUE, &f);
@@ -523,29 +521,44 @@ void replace_buffer(const char *filename)
 	if (descriptor < 0)
 		return;
 
-	/* Reinitialize the text of the current buffer. */
-	free_filestruct(openfile->fileage);
-	initialize_buffer_text();
+#ifndef NANO_TINY
+	add_undo(COUPLE_BEGIN);
+	openfile->undotop->strdata = mallocstrcpy(NULL, _("spelling correction"));
+#endif
+
+	/* Throw away the text of the file. */
+	cutbuffer = NULL;
+	openfile->current = openfile->fileage;
+	openfile->current_x = 0;
+#ifndef NANO_TINY
+	add_undo(CUT_TO_EOF);
+#endif
+	do_cut_text(FALSE, FALSE, TRUE);
+#ifndef NANO_TINY
+	update_undo(CUT_TO_EOF);
+#endif
+	free_filestruct(cutbuffer);
+	cutbuffer = was_cutbuffer;
 
 	/* Insert the processed file into its place. */
-	read_file(f, descriptor, filename, FALSE, TRUE);
+	read_file(f, descriptor, filename, TRUE);
 
-	/* Put current at a place that is certain to exist. */
-	openfile->current = openfile->fileage;
+#ifndef NANO_TINY
+	add_undo(COUPLE_END);
+	openfile->undotop->strdata = mallocstrcpy(NULL, _("spelling correction"));
+#endif
 }
 
 #ifndef NANO_TINY
 /* Open the specified file, and if that succeeds, blow away the text of
- * the current buffer at the given coordinates and read the file
+ * the current buffer covered by the mark and read the file
  * contents into its place. */
-void replace_marked_buffer(const char *filename, filestruct *top, size_t top_x,
-		filestruct *bot, size_t bot_x)
+void replace_marked_buffer(const char *filename)
 {
 	FILE *f;
 	int descriptor;
 	bool old_no_newlines = ISSET(NO_NEWLINES);
-	filestruct *trash_top = NULL;
-	filestruct *trash_bot = NULL;
+	filestruct *was_cutbuffer = cutbuffer;
 
 	descriptor = open_file(filename, FALSE, TRUE, &f);
 
@@ -555,15 +568,26 @@ void replace_marked_buffer(const char *filename, filestruct *top, size_t top_x,
 	/* Don't add a magicline when replacing text in the buffer. */
 	SET(NO_NEWLINES);
 
-	/* Throw away the text under the mark, and insert the processed file
-	 * where the marked text was. */
-	extract_buffer(&trash_top, &trash_bot, top, top_x, bot, bot_x);
-	free_filestruct(trash_top);
-	read_file(f, descriptor, filename, FALSE, TRUE);
+	add_undo(COUPLE_BEGIN);
+	openfile->undotop->strdata = mallocstrcpy(NULL, _("spelling correction"));
+
+	/* Throw away the text under the mark. */
+	cutbuffer = NULL;
+	add_undo(CUT);
+	do_cut_text(FALSE, TRUE, FALSE);
+	update_undo(CUT);
+	free_filestruct(cutbuffer);
+	cutbuffer = was_cutbuffer;
+
+	/* Insert the processed file where the marked text was. */
+	read_file(f, descriptor, filename, TRUE);
 
 	/* Restore the magicline behavior now that we're done fiddling. */
 	if (!old_no_newlines)
 		UNSET(NO_NEWLINES);
+
+	add_undo(COUPLE_END);
+	openfile->undotop->strdata = mallocstrcpy(NULL, _("spelling correction"));
 }
 #endif /* !NANO_TINY */
 #endif /* ENABLE_SPELLER */
@@ -588,12 +612,31 @@ void prepare_for_display(void)
 }
 
 #ifdef ENABLE_MULTIBUFFER
+/* Show name of current buffer and its number of lines on the status bar. */
+void mention_name_and_linecount(void)
+{
+	size_t count = openfile->filebot->lineno -
+						(openfile->filebot->data[0] == '\0' ? 1 : 0);
+#ifndef NANO_TINY
+	if (openfile->fmt != NIX_FILE)
+		/* TRANSLATORS: first %s is file name, second %s is file format. */
+		statusline(HUSH, P_("%s -- %zu line (%s)", "%s -- %zu lines (%s)", count),
+						openfile->filename[0] == '\0' ?
+						_("New Buffer") : tail(openfile->filename), count,
+						openfile->fmt == DOS_FILE ? _("DOS") : _("Mac"));
+	else
+#endif
+		statusline(HUSH, P_("%s -- %zu line", "%s -- %zu lines", count),
+						openfile->filename[0] == '\0' ?
+						_("New Buffer") : tail(openfile->filename), count);
+}
+
 /* Switch to a neighbouring file buffer; to the next if to_next is TRUE;
  * otherwise, to the previous one. */
 void switch_to_adjacent_buffer(bool to_next)
 {
 	/* If only one file buffer is open, say so and get out. */
-	if (openfile == openfile->next && !inhelp) {
+	if (openfile == openfile->next) {
 		statusbar(_("No more open file buffers"));
 		return;
 	}
@@ -618,14 +661,8 @@ void switch_to_adjacent_buffer(bool to_next)
 	/* Ensure that the main loop will redraw the help lines. */
 	currmenu = MMOST;
 
-	/* Indicate the switch on the statusbar. */
-	statusline(HUSH, _("Switched to %s"),
-						((openfile->filename[0] == '\0') ?
-						_("New Buffer") : openfile->filename));
-
-#ifdef DEBUG
-	dump_filestruct(openfile->current);
-#endif
+	/* Indicate on the status bar where we switched to. */
+	mention_name_and_linecount();
 }
 
 /* Switch to the previous entry in the list of open files. */
@@ -716,13 +753,10 @@ char *encode_data(char *buf, size_t buf_len)
 	return mallocstrcpy(NULL, buf);
 }
 
-/* Read an open file into the current buffer.  f should be set to the
- * open file, and filename should be set to the name of the file.
- * undoable means do we want to create undo records to try and undo
- * this.  Will also attempt to check file writability if fd > 0 and
- * checkwritable == TRUE. */
-void read_file(FILE *f, int fd, const char *filename, bool undoable,
-				bool checkwritable)
+/* Read the given open file f into the current buffer.  filename should be
+ * set to the name of the file.  undoable means that undo records should be
+ * created and that the file does not need to be checked for writability. */
+void read_file(FILE *f, int fd, const char *filename, bool undoable)
 {
 	ssize_t was_lineno = openfile->current->lineno;
 		/* The line number where we start the insertion. */
@@ -766,8 +800,12 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable,
 	topline = make_new_node(NULL);
 	bottomline = topline;
 
+	/* Lock the file before starting to read it, to avoid the overhead
+	 * of locking it for each single byte that we read from it. */
+	flockfile(f);
+
 	/* Read the entire file into the new buffer. */
-	while ((input_int = getc(f)) != EOF) {
+	while ((input_int = getc_unlocked(f)) != EOF) {
 		input = (char)input_int;
 
 		/* If it's a *nix file ("\n") or a DOS file ("\r\n"), and file
@@ -834,11 +872,14 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable,
 #endif
 	}
 
+	/* We are done with the file, unlock it. */
+	funlockfile(f);
+
 	/* Perhaps this could use some better handling. */
 	if (ferror(f))
 		nperror(filename);
 	fclose(f);
-	if (fd > 0 && checkwritable) {
+	if (fd > 0 && !undoable) {
 		close(fd);
 		writable = is_file_writable(filename);
 	}
@@ -1042,7 +1083,7 @@ void do_insertfile(void)
 		if (execute) {
 #ifdef ENABLE_MULTIBUFFER
 			if (ISSET(MULTIBUFFER))
-				/* TRANSLATORS: The next four messages are prompts. */
+				/* TRANSLATORS: The next six messages are prompts. */
 				msg = _("Command to execute in new buffer");
 			else
 #endif
@@ -1052,10 +1093,20 @@ void do_insertfile(void)
 		{
 #ifdef ENABLE_MULTIBUFFER
 			if (ISSET(MULTIBUFFER))
-				msg = _("File to insert into new buffer [from %s]");
+#ifndef NANO_TINY
+				if ISSET(NO_CONVERT)
+					msg = _("File to read unconverted into new buffer [from %s]");
+				else
+#endif
+					msg = _("File to read into new buffer [from %s]");
 			else
 #endif
-				msg = _("File to insert [from %s]");
+#ifndef NANO_TINY
+				if ISSET(NO_CONVERT)
+					msg = _("File to insert unconverted [from %s]");
+				else
+#endif
+					msg = _("File to insert [from %s]");
 		}
 
 		present_path = mallocstrcpy(present_path, "./");
@@ -1098,6 +1149,10 @@ void do_insertfile(void)
 			}
 #endif
 #ifndef NANO_TINY
+			if (func == flip_convert) {
+				TOGGLE(NO_CONVERT);
+				continue;
+			}
 			if (func == flip_execute) {
 				execute = !execute;
 				continue;
@@ -1115,6 +1170,13 @@ void do_insertfile(void)
 				answer = chosen;
 				i = 0;
 			}
+#endif
+#ifndef NANO_TINY
+			if (func == flip_pipe) {
+				add_or_remove_pipe_symbol_from_answer();
+				given = mallocstrcpy(given, answer);
+				continue;
+			}
 #endif
 			/* If we don't have a file yet, go back to the prompt. */
 			if (i != 0 && (!ISSET(MULTIBUFFER) || i != -2))
@@ -1125,7 +1187,7 @@ void do_insertfile(void)
 #ifdef ENABLE_MULTIBUFFER
 				/* When in multibuffer mode, first open a blank buffer. */
 				if (ISSET(MULTIBUFFER))
-					open_buffer("", FALSE);
+					open_buffer("", TRUE);
 #endif
 				/* If the command is not empty, execute it and read its output
 				 * into the buffer, and add the command to the history list. */
@@ -1152,8 +1214,8 @@ void do_insertfile(void)
 				/* Make sure the specified path is tilde-expanded. */
 				answer = free_and_assign(answer, real_dir_from_tilde(answer));
 
-				/* Read the specified file into the current buffer. */
-				open_buffer(answer, TRUE);
+				/* Read the file into a new buffer or into current buffer. */
+				open_buffer(answer, ISSET(MULTIBUFFER));
 			}
 
 #ifdef ENABLE_MULTIBUFFER
@@ -1761,6 +1823,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp,
 			umask(original_umask);
 	}
 
+#ifndef NANO_TINY
 	/* If we're prepending, copy the file to a temp file. */
 	if (method == PREPEND) {
 		int fd_source;
@@ -1807,6 +1870,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp,
 			goto cleanup_and_exit;
 		}
 	}
+#endif /* !NANO_TINY */
 
 	if (f_open == NULL) {
 		/* Now open the file in place.  Use O_EXCL if tmp is TRUE.  This
@@ -1888,6 +1952,7 @@ bool write_file(const char *name, FILE *f_open, bool tmp,
 		lineswritten++;
 	}
 
+#ifndef NANO_TINY
 	/* If we're prepending, open the temp file, and append it to f. */
 	if (method == PREPEND) {
 		int fd_source;
@@ -1915,22 +1980,24 @@ bool write_file(const char *name, FILE *f_open, bool tmp,
 		}
 
 		unlink(tempname);
-	} else if (fclose(f) != 0) {
+	} else
+#endif
+	if (fclose(f) != 0) {
 		statusline(ALERT, _("Error writing %s: %s"), realname,
 						strerror(errno));
 		goto cleanup_and_exit;
 	}
 
-	if (method == OVERWRITE && !tmp) {
-		/* If we must set the filename, and it changed, adjust things. */
-		if (fullbuffer && strcmp(openfile->filename, realname) != 0) {
+	/* When having written an entire buffer, update some administrivia. */
+	if (fullbuffer && method == OVERWRITE && !tmp) {
+		/* If the filename was changed, check if this means a new syntax. */
+		if (strcmp(openfile->filename, realname) != 0) {
 #ifdef ENABLE_COLOR
 			const char *oldname, *newname;
 
 			oldname = openfile->syntax ? openfile->syntax->name : "";
 #endif
 			openfile->filename = mallocstrcpy(openfile->filename, realname);
-
 #ifdef ENABLE_COLOR
 			/* See if the applicable syntax has changed. */
 			color_update();
@@ -1947,29 +2014,27 @@ bool write_file(const char *name, FILE *f_open, bool tmp,
 					line->multidata = NULL;
 					line = line->next;
 				}
+
 				precalc_multicolorinfo();
 				refresh_needed = TRUE;
 			}
 #endif
 		}
-
-		if (fullbuffer) {
 #ifndef NANO_TINY
-			/* Get or update the stat info to reflect the current state. */
-			stat_with_alloc(realname, &openfile->current_stat);
+		/* Get or update the stat info to reflect the current state. */
+		stat_with_alloc(realname, &openfile->current_stat);
 
-			/* Record at which point in the undo stack the file was saved. */
-			openfile->last_saved = openfile->current_undo;
-			openfile->last_action = OTHER;
+		/* Record at which point in the undo stack the file was saved. */
+		openfile->last_saved = openfile->current_undo;
+		openfile->last_action = OTHER;
 #endif
-			openfile->modified = FALSE;
-			titlebar(NULL);
-		}
-
-		statusline(HUSH, P_("Wrote %zu line", "Wrote %zu lines",
-						lineswritten), lineswritten);
+		openfile->modified = FALSE;
+		titlebar(NULL);
 	}
 
+	if (!tmp)
+		statusline(HUSH, P_("Wrote %zu line", "Wrote %zu lines",
+								lineswritten), lineswritten);
 	retval = TRUE;
 
   cleanup_and_exit:
@@ -2146,8 +2211,12 @@ int do_writeout(bool exiting, bool withprompt)
 		 * egg.  Display the credits. */
 		if (!did_credits && exiting && !ISSET(TEMP_FILE) &&
 								strcasecmp(answer, "zzy") == 0) {
-			do_credits();
-			did_credits = TRUE;
+			if (LINES > 5 && COLS > 31) {
+				do_credits();
+				did_credits = TRUE;
+			} else
+				/* TRANSLATORS: Concisely say the screen is too small. */
+				statusbar(_("Too tiny"));
 			break;
 		}
 #endif
@@ -2429,8 +2498,8 @@ char **username_tab_completion(const char *buf, size_t *num_matches,
 
 /* We consider the first buf_len characters of buf for filename tab
  * completion. */
-char **cwd_tab_completion(const char *buf, bool allow_files, size_t
-		*num_matches, size_t buf_len)
+char **cwd_tab_completion(const char *buf, bool allow_files,
+		size_t *num_matches, size_t buf_len)
 {
 	char *dirname = mallocstrcpy(NULL, buf);
 	char *slash, *filename;
@@ -2609,7 +2678,8 @@ char *input_tab(char *buf, bool allow_files, size_t *place,
 		if (!*lastwastab)
 			*lastwastab = TRUE;
 		else if (num_matches > 1) {
-			int longest_name = 0, ncols, editline = 0;
+			size_t longest_name = 0, ncols;
+			int editline = 0;
 
 			/* Sort the list of available choices. */
 			qsort(matches, num_matches, sizeof(char *), diralphasort);
