@@ -3,7 +3,7 @@
  *                                                                        *
  *   Copyright (C) 2001-2011, 2013-2018 Free Software Foundation, Inc.    *
  *   Copyright (C) 2014 Mike Frysinger                                    *
- *   Copyright (C) 2014-2017 Benno Schulenberg                            *
+ *   Copyright (C) 2014-2018 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
  *   it under the terms of the GNU General Public License as published    *
@@ -31,14 +31,14 @@
 #ifdef ENABLE_NANORC
 
 #ifndef RCFILE_NAME
+#define HOME_RC_NAME ".nanorc"
 #define RCFILE_NAME "nanorc"
+#else
+#define HOME_RC_NAME RCFILE_NAME
 #endif
 
 static const rcoption rcopts[] = {
 	{"boldtext", BOLD_TEXT},
-#ifdef ENABLE_LINENUMBERS
-	{"linenumbers", LINE_NUMBERS},
-#endif
 #ifdef ENABLE_JUSTIFY
 	{"brackets", 0},
 #endif
@@ -48,6 +48,9 @@ static const rcoption rcopts[] = {
 #endif
 #ifdef ENABLE_HISTORIES
 	{"historylog", HISTORYLOG},
+#endif
+#ifdef ENABLE_LINENUMBERS
+	{"linenumbers", LINE_NUMBERS},
 #endif
 	{"morespace", MORE_SPACE},
 #ifdef ENABLE_MOUSE
@@ -85,16 +88,16 @@ static const rcoption rcopts[] = {
 	{"tempfile", TEMP_FILE},
 	{"view", VIEW_MODE},
 #ifndef NANO_TINY
+	{"afterends", AFTER_ENDS},
 	{"allow_insecure_backup", INSECURE_BACKUP},
 	{"atblanks", AT_BLANKS},
 	{"autoindent", AUTOINDENT},
 	{"backup", BACKUP_FILE},
 	{"backupdir", 0},
-	{"backwards", BACKWARDS_SEARCH},
 	{"casesensitive", CASE_SENSITIVE},
-	{"cut", CUT_FROM_CURSOR},  /* deprecated form, remove in 2020 */
+	{"cut", CUT_FROM_CURSOR},  /* deprecated form, remove end of 2018 */
 	{"cutfromcursor", CUT_FROM_CURSOR},
-	{"justifytrim", TRIM_BLANKS},  /* deprecated form, remove in 2020 */
+	{"justifytrim", TRIM_BLANKS},  /* deprecated form, remove end of 2018 */
 	{"locking", LOCKING},
 	{"matchbrackets", 0},
 	{"noconvert", NO_CONVERT},
@@ -297,7 +300,6 @@ void parse_syntax(char *ptr)
 	live_syntax->headers = NULL;
 	live_syntax->magics = NULL;
 	live_syntax->linter = NULL;
-	live_syntax->formatter = NULL;
 #ifdef ENABLE_COMMENT
 	live_syntax->comment = mallocstrcpy(NULL, GENERAL_COMMENT_CHARACTER);
 #endif
@@ -311,10 +313,6 @@ void parse_syntax(char *ptr)
 
 	opensyntax = TRUE;
 
-#ifdef DEBUG
-	fprintf(stderr, "Starting a new syntax type: \"%s\"\n", nameptr);
-#endif
-
 	/* The default syntax should have no associated extensions. */
 	if (strcmp(live_syntax->name, "default") == 0 && *ptr != '\0') {
 		rcfile_error(N_("The \"default\" syntax does not accept extensions"));
@@ -347,10 +345,6 @@ void parse_binding(char *ptr, bool dobind)
 	sc *s, *newsc = NULL;
 	int menu;
 
-#ifdef DEBUG
-	fprintf(stderr, "Starting the rebinding code...\n");
-#endif
-
 	if (*ptr == '\0') {
 		rcfile_error(N_("Missing key name"));
 		return;
@@ -396,7 +390,8 @@ void parse_binding(char *ptr, bool dobind)
 		if (funcptr[0] == '\0') {
 			rcfile_error(N_("Must specify a function to bind the key to"));
 			goto free_things;
-		}
+		} else if (ptr == NULL)
+			goto free_things;
 	}
 
 	menuptr = ptr;
@@ -419,7 +414,7 @@ void parse_binding(char *ptr, bool dobind)
 			newsc->toggle = 0;
 #endif
 		} else
-		    newsc = strtosc(funcptr);
+			newsc = strtosc(funcptr);
 
 		if (newsc == NULL) {
 			rcfile_error(N_("Cannot map name \"%s\" to a function"), funcptr);
@@ -427,7 +422,7 @@ void parse_binding(char *ptr, bool dobind)
 		}
 	}
 
-	menu = strtomenu(menuptr);
+	menu = name_to_menu(menuptr);
 	if (menu < 1) {
 		rcfile_error(N_("Cannot map name \"%s\" to a menu"), menuptr);
 		goto free_things;
@@ -506,7 +501,7 @@ bool is_good_file(char *file)
 	if (access(file, R_OK) != 0)
 		return FALSE;
 
-	/* If the thing exists, it may not be a directory nor a device. */
+	/* If the thing exists, it may be neither a directory nor a device. */
 	if (stat(file, &rcinfo) != -1 && (S_ISDIR(rcinfo.st_mode) ||
 				S_ISCHR(rcinfo.st_mode) || S_ISBLK(rcinfo.st_mode))) {
 		rcfile_error(S_ISDIR(rcinfo.st_mode) ? _("\"%s\" is a directory") :
@@ -539,10 +534,6 @@ static void parse_one_include(char *file)
 	nanorc = file;
 	lineno = 0;
 
-#ifdef DEBUG
-	fprintf(stderr, "Parsing file \"%s\"\n", file);
-#endif
-
 	parse_rcfile(rcstream, TRUE);
 }
 
@@ -603,10 +594,44 @@ short color_to_short(const char *colorname, bool *bright)
 	else if (strcasecmp(colorname, "black") == 0)
 		return COLOR_BLACK;
 	else if (strcasecmp(colorname, "normal") == 0)
-		return -1;
+		return USE_THE_DEFAULT;
 
 	rcfile_error(N_("Color \"%s\" not understood"), colorname);
-	return -2;
+	return BAD_COLOR;
+}
+
+/* Parse the color name (or pair of color names) in the given string.
+ * Return FALSE when any color name is invalid; otherwise return TRUE. */
+bool parse_color_names(char *combostr, short *fg, short *bg, int *attributes)
+{
+	char *comma = strchr(combostr, ',');
+	bool bright;
+
+	*attributes = A_NORMAL;
+
+	if (comma != NULL) {
+		*bg = color_to_short(comma + 1, &bright);
+		if (bright) {
+			rcfile_error(N_("A background color cannot be bright"));
+			return FALSE;
+		}
+		if (*bg == BAD_COLOR)
+			return FALSE;
+		*comma = '\0';
+	} else
+		*bg = USE_THE_DEFAULT;
+
+	if (comma != combostr) {
+		*fg = color_to_short(combostr, &bright);
+		if (*fg == BAD_COLOR)
+			return FALSE;
+
+		if (bright)
+			*attributes = A_BOLD;
+	} else
+		*fg = USE_THE_DEFAULT;
+
+	return TRUE;
 }
 
 /* Parse the color string in the line at ptr, and add it to the current
@@ -615,7 +640,7 @@ short color_to_short(const char *colorname, bool *bright)
 void parse_colors(char *ptr, int rex_flags)
 {
 	short fg, bg;
-	bool bright;
+	int attributes;
 	char *item;
 
 	if (!opensyntax) {
@@ -631,7 +656,7 @@ void parse_colors(char *ptr, int rex_flags)
 
 	item = ptr;
 	ptr = parse_next_word(ptr);
-	if (!parse_color_names(item, &fg, &bg, &bright))
+	if (!parse_color_names(item, &fg, &bg, &attributes))
 		return;
 
 	if (*ptr == '\0') {
@@ -678,7 +703,7 @@ void parse_colors(char *ptr, int rex_flags)
 
 			newcolor->fg = fg;
 			newcolor->bg = bg;
-			newcolor->bright = bright;
+			newcolor->attributes = attributes;
 			newcolor->rex_flags = rex_flags;
 
 			newcolor->start_regex = mallocstrcpy(NULL, item);
@@ -736,42 +761,12 @@ void parse_colors(char *ptr, int rex_flags)
 	}
 }
 
-/* Parse the color name, or pair of color names, in combostr. */
-bool parse_color_names(char *combostr, short *fg, short *bg, bool *bright)
-{
-	char *comma = strchr(combostr, ',');
-
-	if (comma != NULL) {
-		*bg = color_to_short(comma + 1, bright);
-		if (*bright) {
-			rcfile_error(N_("A background color cannot be bright"));
-			return FALSE;
-		}
-		*comma = '\0';
-	} else
-		*bg = -1;
-
-	if (comma != combostr) {
-		*fg = color_to_short(combostr, bright);
-
-		/* If the specified foreground color is bad, ignore the regexes. */
-		if (*fg == -2)
-			return FALSE;
-	} else
-		*fg = -1;
-
-	if (*bg == -2)
-		*bg = -1;
-
-	return TRUE;
-}
-
 /* Parse the argument of an interface color option. */
 colortype *parse_interface_color(char *combostr)
 {
 	colortype *trio = nmalloc(sizeof(colortype));
 
-	if (parse_color_names(combostr, &trio->fg, &trio->bg, &trio->bright)) {
+	if (parse_color_names(combostr, &trio->fg, &trio->bg, &trio->attributes)) {
 		free(combostr);
 		return trio;
 	} else {
@@ -844,7 +839,7 @@ void grab_and_store(const char *kind, char *ptr, regexlisttype **storage)
 	}
 }
 
-/* Gather and store the string after a comment/linter/formatter command. */
+/* Gather and store the string after a comment/linter command. */
 void pick_up_name(const char *kind, char *ptr, char **storage)
 {
 	if (!opensyntax) {
@@ -883,17 +878,18 @@ static void check_vitals_mapped(void)
 {
 	subnfunc *f;
 	int v;
-#define VITALS 5
-	void (*vitals[VITALS])(void) = { do_exit, do_exit, do_cancel, do_cancel, do_cancel };
-	int inmenus[VITALS] = { MMAIN, MHELP, MWHEREIS, MREPLACE, MGOTOLINE };
+#define VITALS 3
+	void (*vitals[VITALS])(void) = { do_exit, do_exit, do_cancel };
+	int inmenus[VITALS] = { MMAIN, MHELP, MYESNO };
 
 	for  (v = 0; v < VITALS; v++) {
 		for (f = allfuncs; f != NULL; f = f->next) {
 			if (f->func == vitals[v] && f->menus & inmenus[v]) {
 				const sc *s = first_sc_for(inmenus[v], f->func);
 				if (!s) {
-					fprintf(stderr, _("Fatal error: no keys mapped for function "
-										"\"%s\".  Exiting.\n"), f->desc);
+					fprintf(stderr, _("No key is bound to function '%s' in "
+										"menu '%s'.  Exiting.\n"), f->desc,
+										menu_to_name(inmenus[v]));
 					fprintf(stderr, _("If needed, use nano with the -I option "
 										"to adjust your nanorc settings.\n"));
 					exit(1);
@@ -938,7 +934,7 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 
 #ifdef ENABLE_COLOR
 		/* Handle extending first... */
-		if (strcasecmp(keyword, "extendsyntax") == 0) {
+		if (strcasecmp(keyword, "extendsyntax") == 0 && !syntax_only) {
 			syntaxtype *sint;
 			char *syntaxname = ptr;
 
@@ -995,13 +991,12 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 			parse_colors(ptr, NANO_REG_EXTENDED | REG_ICASE);
 		else if (strcasecmp(keyword, "linter") == 0)
 			pick_up_name("linter", ptr, &live_syntax->linter);
-		else if (strcasecmp(keyword, "formatter") == 0)
-#ifdef ENABLE_SPELLER
-			pick_up_name("formatter", ptr, &live_syntax->formatter);
-#else
-			;
-#endif
-		else if (syntax_only)
+		else if (syntax_only && (strcasecmp(keyword, "set") == 0 ||
+								strcasecmp(keyword, "unset") == 0 ||
+								strcasecmp(keyword, "bind") == 0 ||
+								strcasecmp(keyword, "unbind") == 0 ||
+								strcasecmp(keyword, "include") == 0 ||
+								strcasecmp(keyword, "extendsyntax") == 0))
 			rcfile_error(N_("Command \"%s\" not allowed in included file"),
 										keyword);
 		else if (strcasecmp(keyword, "include") == 0)
@@ -1047,10 +1042,6 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 			continue;
 		}
 
-#ifdef DEBUG
-		fprintf(stderr, "    Option name = \"%s\"\n", rcopts[i].name);
-		fprintf(stderr, "    Flag = %ld\n", rcopts[i].flag);
-#endif
 		/* First handle unsetting. */
 		if (set == -1) {
 			if (rcopts[i].flag != 0)
@@ -1079,15 +1070,14 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 		ptr = parse_argument(ptr);
 
 		option = mallocstrcpy(NULL, option);
-#ifdef DEBUG
-		fprintf(stderr, "    Option argument = \"%s\"\n", option);
-#endif
-		/* Make sure the option argument is a valid multibyte string. */
-		if (!is_valid_mbstring(option)) {
+
+#ifdef ENABLE_UTF8
+		/* When in a UTF-8 locale, ignore arguments with invalid sequences. */
+		if (using_utf8() && mbstowcs(NULL, option, 0) == (size_t)-1) {
 			rcfile_error(N_("Argument is not a valid multibyte string"));
 			continue;
 		}
-
+#endif
 #ifdef ENABLE_COLOR
 		if (strcasecmp(rcopts[i].name, "titlecolor") == 0)
 			color_combo[TITLE_BAR] = parse_interface_color(option);
@@ -1112,10 +1102,10 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 #endif
 #ifdef ENABLED_WRAPORJUSTIFY
 		if (strcasecmp(rcopts[i].name, "fill") == 0) {
-			if (!parse_num(option, &wrap_at)) {
+			if (!parse_num(option, &fill)) {
 				rcfile_error(N_("Requested fill size \"%s\" is invalid"),
 								option);
-				wrap_at = -CHARS_FROM_EOL;
+				fill = -COLUMNS_FROM_EOL;
 			} else
 				UNSET(NO_WRAP);
 			free(option);
@@ -1144,19 +1134,17 @@ void parse_rcfile(FILE *rcstream, bool syntax_only)
 #endif
 #ifdef ENABLE_JUSTIFY
 		if (strcasecmp(rcopts[i].name, "punct") == 0) {
-			punct = option;
-			if (has_blank_mbchars(punct)) {
+			if (has_blank_mbchars(option)) {
 				rcfile_error(N_("Non-blank characters required"));
-				free(punct);
-				punct = NULL;
-			}
+				free(option);
+			} else
+				punct = option;
 		} else if (strcasecmp(rcopts[i].name, "brackets") == 0) {
-			brackets = option;
-			if (has_blank_mbchars(brackets)) {
+			if (has_blank_mbchars(option)) {
 				rcfile_error(N_("Non-blank characters required"));
-				free(brackets);
-				brackets = NULL;
-			}
+				free(option);
+			} else
+				brackets = option;
 		} else if (strcasecmp(rcopts[i].name, "quotestr") == 0)
 			quotestr = option;
 		else
@@ -1208,9 +1196,6 @@ void parse_one_nanorc(void)
 	if (!is_good_file(nanorc))
 		return;
 
-#ifdef DEBUG
-	fprintf(stderr, "Going to parse file \"%s\"\n", nanorc);
-#endif
 	rcstream = fopen(nanorc, "rb");
 
 	/* If opening the file succeeded, parse it.  Otherwise, only
@@ -1237,9 +1222,8 @@ void do_rcfiles(void)
 {
 	const char *xdgconfdir;
 
+	/* First process the system-wide nanorc, if there is one. */
 	nanorc = mallocstrcpy(nanorc, SYSCONFDIR "/nanorc");
-
-	/* Process the system-wide nanorc. */
 	parse_one_nanorc();
 
 	/* When configured with --disable-wrapping-as-root, turn wrapping off
@@ -1252,9 +1236,9 @@ void do_rcfiles(void)
 	get_homedir();
 	xdgconfdir = getenv("XDG_CONFIG_HOME");
 
-	/* Now try the to find a nanorc file in the user's home directory
-	 * or in the XDG configuration directories. */
-	if (have_nanorc(homedir, "/." RCFILE_NAME))
+	/* Now try the to find a nanorc file in the user's home directory or in
+	 * the XDG configuration directories, and process the first one found. */
+	if (have_nanorc(homedir, "/" HOME_RC_NAME))
 		parse_one_nanorc();
 	else if (have_nanorc(xdgconfdir, "/nano/" RCFILE_NAME))
 		parse_one_nanorc();
@@ -1265,6 +1249,16 @@ void do_rcfiles(void)
 
 	check_vitals_mapped();
 
+#ifdef __linux__
+	/* On a Linux console, don't start nano when there are rcfile errors,
+	 * because otherwise these error messages get wiped. */
+	if (on_a_vt && rcfile_with_errors) {
+		fprintf(stderr, _("If needed, use nano with the -I option "
+							"to adjust your nanorc settings.\n"));
+		exit(1);
+	}
+#endif
+
 	free(nanorc);
 }
 
